
// Written by Hans de Nivelle, November 2014. 

#include "airfoil_mach0.h"

#include <cstdlib>
#include <math.h> 


namespace
{
   std::map< double, double > 
   getlift( const std::vector< coefficients > & table )
   {
      std::map< double, double > result;
      for( const auto& t: table ) result [ t. alpha ] = t. lift;
      return result; 
   }

   std::map< double, double > 
   getdrag( const std::vector< coefficients > & table )
   {
      std::map< double, double > result;
      for( const auto& t: table ) result [ t. alpha ] = t. drag;
      return result;
   }

   std::map< double, double > 
   getmoment( const std::vector< coefficients > & table )
   {
      std::map< double, double > result;
      for( const auto& t: table ) result [ t. alpha ] = t. moment;
      return result;
   }
}


airfoil_mach0::airfoil_mach0( const std::string& name,
                              double alpha0, double alpha1,
                              const std::vector< coefficients > & positions,
                              double deriv2begin, double deriv2end, 
                              math::interpolator_name choice )
   : name{ name },
     alpha0{ alpha0 }, alpha1{ alpha1 }, 
     lift{ math::interpolator::choose( choice, getlift( positions ), 
                                       deriv2begin, deriv2end ) },
     drag{ math::interpolator::choose( choice, getdrag( positions ), 
                                       deriv2begin, deriv2end ) },
     moment{ math::interpolator::choose( choice, getmoment( positions ), 
                                         deriv2begin, deriv2end ) } 
{
}

coefficients airfoil_mach0::zeroliftpoint( ) const
{
   // A simple, binary search:

   double lowerbound = alpha0;
   double upperbound = alpha1;

   coefficients c = getcoefficients( lowerbound );
   while( upperbound - lowerbound > 1.0E-5 )
   {
      double middle = ( lowerbound + upperbound ) / 2;
      c = getcoefficients( middle );
      if( c. lift > 0.0 )
         upperbound = middle;
      else
         lowerbound = middle;
   }
   return c;    
}

coefficients airfoil_mach0::mindragpoint( ) const
{
   coefficients c = getcoefficients( alpha0 );
   for( double d = alpha0; d <= alpha1; d += ( alpha1 - alpha0 ) * 1.0E-5 )
   {
      auto cc = getcoefficients(d);
      if( cc. drag < c. drag )
         c = cc;
   }
   return c;
}

///////////////////////////////////////////////////////////////////

airfoil_mach0 airfoils::raf15( )
{
   return airfoil_mach0( "RAF 15", -6.0_deg, 22.0_deg,
   {
      { -4.0_deg,  -0.14,   0.014,  0,   -0.036 },
      { -2.0_deg,   0.02,   0.008,  0,   -0.052 },
      {  0.0_deg,   0.14,   0.008,  0,   -0.090 },
      {  2.0_deg,   0.32,   0.012,  0,   -0.130 },
      {  4.0_deg,   0.46,   0.020,  0,   -0.160 },
      {  6.0_deg,   0.60,   0.030,  0,   -0.200 },
      {  8.0_deg,   0.76,   0.044,  0,   -0.240 },
      { 10.0_deg,   0.90,   0.060,  0,   -0.280 },
      { 12.0_deg,   1.04,   0.070,  0,   -0.310 },
      { 14.0_deg,   1.16,   0.096,  0,   -0.330 },
      { 15.0_deg,   1.22,   0.110,  0,   -0.340 },
      { 16.0_deg,   1.16,   0.140,  0,   -0.350 },  
      { 18.0_deg,   1.02,   0.210,  0,   -0.384 },
      { 20.0_deg,   0.94,   0.260,  0,   -0.390 }
   }, 0, 0, math::interpolator_name::spline );
}

airfoil_mach0 airfoils::raf15withslot( )
{
   // Lists position of pressure center:

   return airfoil_mach0( "RAF 15 with slot", -6.0_deg, 36.0_deg,
   {
      { -4.0_deg,  -0.27,   0.120,  0.7 },
      { -2.0_deg,  -0.12,   0.075,  0.65 },
      {  0.0_deg,   0.03,   0.056,  0.6 },
      {  2.0_deg,   0.18,   0.050,  0.70 },
      {  4.0_deg,   0.33,   0.050,  0.45 },
      {  6.0_deg,   0.47,   0.053,  0.39 },
      {  8.0_deg,   0.62,   0.060,  0.37 },
      { 10.0_deg,   0.76,   0.070,  0.36 },
      { 12.0_deg,   0.90,   0.080,  0.34 },
      { 14.0_deg,   1.05,   0.100,  0.32 },
      { 15.0_deg,   1.12,   0.110,  0.31 },
      { 16.0_deg,   1.21,   0.120,  0.31 },
      { 18.0_deg,   1.36,   0.148,  0.30 },
      { 20.0_deg,   1.51,   0.175,  0.30 },
      { 24.0_deg,   1.70,   0.234,  0.30 },
      { 26.0_deg,   1.76,   0.270,  0.30 },
      { 28.0_deg,   1.76,   0.304,  0.30 },
      { 30.0_deg,   1.64,   0.344,  0.30 },
      { 34.0_deg,   1.30,   0.430,  0.30 }
   }, 0, 0, math::interpolator_name::spline );
}
      
airfoil_mach0 airfoils::naca0009( )
{
   return airfoil_mach0( "NACA 0009 without flap", -16.0_deg, 16.0_deg,
   {
      { -14.0_deg, -0.87,     0.036,    0.25,     0.012    },
      { -12.0_deg, -0.89,     0.028,    0.25,     0.004    },
      { -10.0_deg, -0.90,     0.021,    0.25,     0.002    },
      { -8.0_deg,  -0.85,     0.018,    0.25,     0.0      },
      { -6.0_deg,  -0.64,     0.014,    0.25,     0.0      },
      { -4.0_deg,  -0.43,     0.011,    0.25,     0.0      },
      { -2.0_deg,  -0.21,     0.010,    0.25,     0.0      },
      { 0.0_deg,    0.0,      0.009,    0.25,     0.0      },
      { 2.0_deg,    0.21,     0.010,    0.25,     0.0      },
      { 4.0_deg,    0.43,     0.011,    0.25,     0.0      },
      { 6.0_deg,    0.64,     0.014,    0.25,     0.0      },
      { 8.0_deg,    0.85,     0.018,    0.25,     0.0      },
      { 10.0_deg,   0.90,     0.021,    0.25,    -0.002    },
      { 12.0_deg,   0.89,     0.028,    0.25,    -0.004    },
      { 14.0_deg,   0.87,     0.036,    0.25,    -0.012    }
   }, 0, 0, math::interpolator_name::spline );
}

airfoil_mach0 airfoils::naca4412( )
{
   double ac = 0.246;

   return airfoil_mach0( "NACA 4412 without flap", -10.0_deg, 18.0_deg, 
   {
      { -8.0_deg,    -0.45,    0.022,  ac,   -0.097 },
      { -6.0_deg,    -0.23,    0.014,  ac,   -0.092 },
      { -4.0_deg,    -0.03,    0.012,  ac,   -0.092 },
      { -2.0_deg,     0.20,    0.010,  ac,   -0.092 },
      {  0.0_deg,     0.38,    0.010,  ac,   -0.093 },
      {  2.0_deg,     0.60,    0.010,  ac,   -0.095 },
      {  4.0_deg,     0.80,    0.012,  ac,   -0.098 },
      {  6.0_deg,     1.00,    0.014,  ac,   -0.100 },
      {  8.0_deg,     1.15,    0.017,  ac,   -0.100 },
      {  10.0_deg,    1.27,    0.022,  ac,   -0.095 },
      {  12.0_deg,    1.36,    0.030,  ac,   -0.092 },
      {  14.0_deg,    1.35,    0.042,  ac,   -0.092 },
      {  16.0_deg,    1.25,    0.059,  ac,   -0.095 }
   }, 0, 0, math::interpolator_name::spline );
}

airfoil_mach0 airfoils::naca4412withflap( )
{
   double ac = 0.25;

   return airfoil_mach0( "NACA 4412 with flap", -10.0_deg, 14.0_deg,
   {
      { -8.0_deg,     0.90,    0.022,  ac,   -0.287 },
      { -6.0_deg,     1.12,    0.014,  ac,   -0.297 },
      { -4.0_deg,     1.34,    0.012,  ac,   -0.302 },
      { -2.0_deg,     1.56,    0.010,  ac,   -0.305 },
      {  0.0_deg,     1.75,    0.010,  ac,   -0.305 },
      {  2.0_deg,     1.95,    0.010,  ac,   -0.305 },
      {  4.0_deg,     2.14,    0.012,  ac,   -0.305 },
      {  6.0_deg,     2.43,    0.014,  ac,   -0.302 },
      {  8.0_deg,     2.50,    0.017,  ac,   -0.300 },
      {  10.0_deg,    2.65,    0.022,  ac,   -0.290 },
      {  12.0_deg,    2.63,    0.030,  ac,   -0.275 } 
   }, 0, 0, math::interpolator_name::spline );
}


airfoil_mach0 airfoils::clarkyh( )
{
   // We list pitching moment coefficients around leading edge:
   
   return airfoil_mach0( "CLARK YH", -6.0_deg, 32.0_deg, 
   {
       {  -4.0_deg,   -0.09,        0.010,     0,    0.030 },
       {  -2.0_deg,    0.05,        0.009,     0,   -0.010 },
       {   0.0_deg,    0.20,        0.010,     0,   -0.046 },
       {   2.0_deg,    0.36,        0.015,     0,   -0.072 },
       {   4.0_deg,    0.51,        0.022,     0,   -0.116 },
       {   6.0_deg,    0.66,        0.033,     0,   -0.150 },
       {   8.0_deg,    0.80,        0.045,     0,   -0.184 },
       {  10.0_deg,    0.94,        0.062,     0,   -0.220 },
       {  12.0_deg,    1.06,        0.083,     0,   -0.244 },
       {  14.0_deg,    1.21,        0.103,     0,   -0.276 },
       {  16.0_deg,    1.33,        0.125,     0,   -0.320 },
       {  18.0_deg,    1.43,        0.146,     0,   -0.352 },
       {  19.0_deg,    1.36,        0.170,     0,   -0.356 },
       {  20.0_deg,    1.26,        0.211,     0,   -0.354 },
       {  25.0_deg,    0.97,        0.324,     0,   -0.354 },
       {  30.0_deg,    0.81,        0.430,     0,   -0.352 }
   }, 0, 0, math::interpolator_name::spline );
}


airfoil_mach0 airfoils::fuselage( )
{

   // We assume a lot:
   // A fuselage is a bad wing. It has C_L half of usual wing,
   // and it stalls at 10 deg.
   // C_D is 0.05 + C_L^2 

   double c = 0.5;
   double ac = 0.25;

   return airfoil_mach0( "Fuselage", -180.0_deg, 180.0_deg, 
   {
     
      {-180.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-180.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-180.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-180.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-170.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-160.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-150.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-140.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-130.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-120.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-110.0_deg,  0.0,       0.5,                         0.5, 0 },
      {-100.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -90.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -80.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -70.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -60.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -50.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -40.0_deg,  0.0,       0.5,                         0.5, 0 },
      { -30.0_deg,  0.0,       0.5,                         0.5, 0 }, 
      { -25.0_deg,  0.0,       0.5,                         0.5, 0 },

      { -20.0_deg, -0.1 / 6,   0.05 + c * ( 0.1 * 0.1 ),    ac, 0 },  
      { -15.0_deg, -0.3 / 6,   0.05 + c * ( 0.3 * 0.3 ),    ac, 0 },
      { -10.0_deg, -0.4 / 6,   0.05 + c * ( 0.4 * 0.4 ),    ac, 0 },
      {  -8.0_deg, -0.32 / 6,  0.05 + c * ( 0.32 * 0.32 ),  ac, 0 },
      {  -6.0_deg, -0.24 / 6,  0.05 + c * ( 0.24 * 0.24 ),  ac, 0 },
      {  -4.0_deg, -0.16 / 6,  0.05 + c * ( 0.16 * 0.16 ),  ac, 0 },
      {  -2.0_deg, -0.08 / 6,  0.05 + c * ( 0.08 * 0.08 ),  ac, 0 },
      {   0.0_deg, 0.0 / 6,    0.05 + c * ( 0.0 * 0.0 ),    ac, 0 },
      {   2.0_deg, 0.08 / 6,   0.05 + c * ( 0.08 * 0.08 ),  ac, 0 },
      {   4.0_deg, 0.16 / 6,   0.05 + c * ( 0.16 * 0.16 ),  ac, 0 },
      {   6.0_deg, 0.24 / 6,   0.05 + c * ( 0.24 * 0.24 ),  ac, 0 },
      {   8.0_deg, 0.32 / 6,   0.05 + c * ( 0.32 * 0.32 ),  ac, 0 },
      {  10.0_deg, 0.4 / 6,    0.05 + c * ( 0.4 * 0.4 ),    ac, 0 },
      {  15.0_deg, 0.3 / 6,    0.05 + c * ( 0.3 * 0.3 ),    ac, 0 },
      {  20.0_deg, 0.1 / 6,    0.05 + c * ( 0.1 * 0.1 ),    ac, 0 },

      {  25.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  30.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  40.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  50.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  60.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  70.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  80.0_deg,  0.0,       0.5,                         0.5, 0 },
      {  90.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 100.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 110.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 120.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 130.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 140.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 150.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 160.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 170.0_deg,  0.0,       0.5,                         0.5, 0 },
      { 180.0_deg,  0.0,       0.5,                         0.5, 0 }

   }, 0,0, math::interpolator_name::spline ); 
}


// The C_L and C_D values above 14.0 degrees come from 
// http://www.aerospaceweb.org/question/aerodynamics/q0194.shtml
// The values below 14.0 degrees come from NACA 0009. 
// The moment coefficients are mostly the result of my phantasy.
// It looks reasonable, when plotted.

airfoil_mach0 airfoils::full_range( )
{
   return airfoil_mach0( "Full Range Data", 0.0_deg, 180.0_deg,
   {
      {  -10.0_deg, -0.9,      0.021,    0.25,   0 },
      {  -8.0_deg,  -0.85,     0.018,    0.25,   0 },
      {  -6.0_deg,  -0.64,     0.014,    0.25,   0 },
      {  -4.0_deg,  -0.43,     0.011,    0.25,   0 },
      {  -2.0_deg,  -0.21,     0.010,    0.25,   0 },

      {  0.0_deg,   0.0,      0.009,    0.25,   0 },

      {  2.0_deg,   0.21,     0.010,    0.25,   0 },
      {  4.0_deg,   0.43,     0.011,    0.25,   0 },
      {  6.0_deg,   0.64,     0.014,    0.25,   0 },
      {  8.0_deg,   0.85,     0.018,    0.25,   0 },
      {  10.0_deg,  0.9,      0.021,    0.25,   0 },
      {  12.0_deg,  0.89,     0.028,    0.25,   0 },
      {  14.0_deg,  0.87,     0.036,    0.25,   0 },
      {  17.0_deg,  0.5802,   0.2164,   0.25,   0 }, 
      {  23.0_deg,  0.5752,   0.3378,   0.25,   0 },
      {  33.0_deg,  0.8826,   0.6971,   0.5,    0 },    
      {  45.0_deg,  0.9620,   1.0828,   0.5,    0 },
      {  55.0_deg,  0.8579,   1.4210,   0.5,    0 },
      {  70.0_deg,  0.5603,   1.6585,   0.5,    0 },
      {  80.0_deg,  0.3273,   1.8010,   0.5,    0 },
      {  90.0_deg,  0.0744,   1.8377,   0.5,    0 },
      { 100.0_deg, -0.1835,   1.7581,   0.5,    0 },
      { 110.0_deg, -0.4265,   1.6362,   0.5,    0 },
      { 120.0_deg, -0.6298,   1.5037,   0.5,    0 },
      { 130.0_deg, -0.8132,   1.2601,   0.5,    0 },
      { 140.0_deg, -0.8975,   0.9425,   0.5,    0 },
      { 150.0_deg, -0.7041,   0.6038,   0.6,    0 },
      { 160.0_deg, -0.5802,   0.3126,   0.75,   0 },
      { 170.0_deg, -0.8132,   0.1325,   0.75,   0 },
      { 180.0_deg,  0.0000,   0.009,    0.75,   0 }

   }, 0, 0, math::interpolator_name::spline );
}



