
#ifndef AIRFOIL_MACH0_INCLUDED
#define AIRFOIL_MACH0_INCLUDED   1

// airfoil_mach0 contains three interpolators, one for
// each of the coefficents C_L, C_D, C_M.
//
// It is assumed that these are the coefficients at low Mach number
// (Mach 0). 


#include <vector> 
#include <cmath>
#include "math/interpolator.h"
#include "coefficients.h"

struct airfoil_mach0
{
   std::string name;

   // alpha0 .. alpha1 is the range for which we have meaningful values.
   // Angle of attack is always in radians!

   double alpha0; 
   double alpha1; 

   math::interpolator lift;
   math::interpolator drag;
   math::interpolator moment;

   airfoil_mach0( const std::string& name,
                  double alpha0, double alpha1,
                  const std::vector< coefficients > & entries,  
                  double deriv2begin, double deriv2end, 
                  math::interpolator_name = math::interpolator_name::spline );
      // One should always plot and inspect the tables.

   coefficients getcoefficients( double alpha ) const 
   {
      return coefficients( alpha,
                           lift( alpha ), drag( alpha ), 0.0, moment( alpha ));
   }

   coefficients operator( ) ( double alpha ) const 
   {
      return coefficients( alpha, 
                           lift( alpha ), drag( alpha ), 0.0, moment( alpha ));
   }


   coefficients zeroliftpoint( ) const;
      // Returns coefficient in which lift coefficient is zero.

   coefficients mindragpoint( ) const;
      // Returns coefficient in which drag coefficient is minimal.
      // This incredibly inefficient function should be used only 
      // during development, not during actual flight. 

};


namespace airfoils
{
   airfoil_mach0 raf15( );
   airfoil_mach0 raf15withslot( );
      // From page 415 of AC Kermode.
   
   airfoil_mach0 clarkyh( ); 
      // From page 416-417 of AC Kermode.

   airfoil_mach0 naca0009( );  // I have shortage of such tables. 
   airfoil_mach0 naca4412( );
   airfoil_mach0 naca4412withflap( );
      // From page 420-421 of AC Kermode.
      // The profile nace4412withflap is useless, because it has no
      // C_D values. The current values are based on my phantasy. 

   airfoil_mach0 clarkyh1( );
      // From page 419 of AC Kermode. The first one uses
      // pitching moment. The second one uses centre of pressure. 
      // They should be the same but they are not. 

   airfoil_mach0 fuselage( );
      // Lots of speculation here. 

   airfoil_mach0 full_range( );
      // A default airfoil that is defined over the full range
      // from -180.0_deg to +180.0_deg. 
}

std::ostream& operator << ( std::ostream& stream, const airfoil_mach0& ); 

#endif

