

#include "graphics.h" 
#include "linalg/affinemap.h"
#include <stdexcept>

#include <SFML/Window.hpp>
#include <SFML/OpenGL.hpp>

void graphics::setview( const linalg::fourmatrix& perspective,
                        const linalg::rototrans& position )
{
   GLdouble mat_flattened [16];
      // OpenGL needs the matrix flattened.

   // Perspective goes into PROJECTION:

   GLdouble* p = mat_flattened; 
   p << perspective;         
   if( p != mat_flattened + 16 )
      throw std::runtime_error( "length does not fit" ); 

   glMatrixMode( GL_PROJECTION );
   glLoadMatrixd( mat_flattened );

   // Inverse of position goes into MODELVIEW:

   p = mat_flattened;
   p << linalg::fourmatrix( position. inverse( ). getaffinemap ( )); 
   if( p != mat_flattened + 16 )
      throw std::runtime_error( "length does not fit" );
   
   glMatrixMode( GL_MODELVIEW );
   glLoadMatrixd( mat_flattened );
}


void graphics::setlight0( linalg::vector lightposition, 
                          bool inf, 
                          scenery::color ambient,
                          scenery::color diffuse,
                          scenery::color specular )
{
   glEnable( GL_LIGHT0 );
 
   static GLfloat vect [4]; 

   vect [0] = lightposition. x;
   vect [1] = lightposition. y;
   vect [2] = lightposition. z;
   vect [3] = inf ? 0.0 : 1.0;   
   glLightfv( GL_LIGHT0, GL_POSITION, vect );

   if( inf )
   {
      glLightf( GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0 );
      glLightf( GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0 );
      glLightf( GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0 );
   }
   else
   {
      glLightf( GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0 );
      glLightf( GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0 );
      glLightf( GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 1.0E-5 );
   }

   vect [0] = ambient. red;
   vect [1] = ambient. green;
   vect [2] = ambient. blue;
   vect [3] = 1.0;
   glLightfv( GL_LIGHT0, GL_AMBIENT, vect );

   vect [0] = diffuse. red;
   vect [1] = diffuse. green;
   vect [2] = diffuse. blue;
   vect [3] = 1.0;
   glLightfv( GL_LIGHT0, GL_DIFFUSE, vect ); 

   vect [0] = specular. red;
   vect [1] = specular. green;
   vect [2] = specular. blue;
   vect [3] = 1.0;
   glLightfv( GL_LIGHT0, GL_SPECULAR, vect );
}

void graphics::setfog( float density, scenery::color col )
{
   static GLfloat vect [4];

   vect[0] = col. red;
   vect[1] = col. green;
   vect[2] = col. blue;
   vect[3] = 1.0;

   glFogfv( GL_FOG_COLOR, vect );
   glFogf( GL_FOG_DENSITY, density );

   glFogf( GL_FOG_START, 5.0 );
   glFogf( GL_FOG_END, 1000.0 );
}
 
void graphics::preparefor3d( scenery::color skycolor )
{
   // Clear screen with skycolor. 

   glClearColor( skycolor. red, skycolor. green, skycolor. blue, 1.0 );
   glClearDepth( 1.0 );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // Enable the depth test, using less-or-equal. 

   glEnable( GL_DEPTH_TEST );
   glDepthFunc( GL_LEQUAL );

   // Allow use of offsets. 

   glEnable( GL_POLYGON_OFFSET_POINT );
   glEnable( GL_POLYGON_OFFSET_LINE );
   glEnable( GL_POLYGON_OFFSET_FILL );

   // Allow continues shading:

   glShadeModel( GL_SMOOTH );

   glFrontFace( GL_CCW );
   glCullFace( GL_BACK );

   glEnable( GL_FOG );
   glFogf( GL_FOG_DENSITY, 0.0 );
   glFogi( GL_FOG_MODE, GL_EXP );

   glEnable( GL_LIGHTING );

   // Blending:

   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

}

void graphics::preparefor2d( )
{
   glClearColor( 0.0, 0.0, 0.0, 1.0 );
   glClearDepth( 1.0 ); 
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   glEnable( GL_DEPTH_TEST );
   glDepthFunc( GL_LEQUAL );
   glDepthMask( GL_TRUE );        // Modifying depth buffer is allowed. 
 
   glShadeModel( GL_SMOOTH );

   glDisable( GL_FOG );
   glDisable( GL_LIGHTING );

   glDisable( GL_BLEND );
}
 
void graphics::drawgrid( double x0, double x1, double y0, double y1 )
{
   if( x0 >= x1 ) 
      throw std::runtime_error( "drawgrid: requires x0 < x1" );
   if( y0 >= y1 ) 
      throw std::runtime_error( "drawgrid: requires y0 < y1" );

   int xlog = floor( log10( x1 - x0 ) - 0.75 );
   double xres = pow( 10, xlog );
   int xstart = floor( x0 / xres );
   int xend = ceil( x1 / xres );

   int ylog = floor( log10( y1 - y0 ) - 0.75 );
   double yres = pow( 10, ylog ); 
   int ystart = floor( y0 / yres );
   int yend = ceil( y1 / yres );

   for( int xi = xstart; xi < xend; ++ xi )
   {
      if( xi == 0 )
         glColor3f( 1,1,1 );
      else
      {
         if( ( xi % 10 == 0 ) == ( xlog & 1 ))
            glColor3f( 0, 0.4, 0 );
         else
            glColor3f( 0.4, 0, 0 );
      }

      double x = xi * xres;

      glBegin( GL_LINE_STRIP );
      glVertex2d( x, y0 );
      glVertex2d( x, y1 ); 
      glEnd( ); 
   }


   for( int yi = ystart; yi < yend; ++ yi )
   {
      if( yi == 0 )
         glColor3f( 1,1,1 );
      else
      {
         if( ( yi % 10 == 0 ) == ( ylog & 1 ))
            glColor3f( 0, 0.4, 0 );
         else
            glColor3f( 0.4, 0, 0 );
      } 

      double y = yi * yres;
 
      glBegin( GL_LINE_STRIP );
      glVertex2d( x0, y );
      glVertex2d( x1, y );
      glEnd( );
   }

   glEnd( );
   glFlush( ); 
}


// Plot a function table. We draw straight lines between the points:


void graphics::plot( const std::function< double(double) > & f, 
                     double x0, double x1, scenery::color col ) 
{
   unsigned int k = 2000;

   double step = ( x1 - x0 ) / k;

   glColor3f( col.red, col.green, col.blue );

   glBegin( GL_LINE_STRIP );
 
   for( unsigned int i = 0; i <= k; ++ i )
   {
      double x = x0 + i * step; 
      glVertex2d( x, f(x) );
   }

   glEnd( );
   glFlush( );
}

