
// Written by Hans de Nivelle, December 2010. 

// fourvector is an implementation of homogeneous coordinates.
// We hardly use them, only in the interface to OpenGL. 
// It is better to use affinemap. 

#ifndef LINALG_FOURVECTOR_INCLUDED
#define LINALG_FOURVECTOR_INCLUDED   1

#include <iostream>
#include "vector.h"
#include <stdexcept>

namespace linalg 
{

   struct fourvector
   {
      double x;
      double y;
      double z;
      double w; 

      fourvector( ) = delete;
      
      fourvector( double x, double y, double z, double w )
         : x{x}, y{y}, z{z}, w{w}
      { }


      fourvector( double x, double y, double z )
         : x{x}, y{y}, z{z}, w{ 1.0 }
      { }       


      fourvector( const vector& v )
         : x{ v.x }, y{ v.y }, z{ v.z }, w{ 1.0 }
      { }


      vector get3vector( ) const
      {
         // I added this check, because OpenGL doesn't seem to handle 
         // vectors with negative w correctly. 
         // (and one should not divide by zero of course.) 

         if( w <= 0.0 )
            throw std::runtime_error( "4-vector has w-component <= 0" );

         return vector( x / w, y / w, z / w );
      }
 
   };


   inline fourvector operator * ( double d, const fourvector& v )
   {
      return fourvector( d * v.x, d * v.y, d * v.z, v. w );
   }

   inline fourvector operator * ( const fourvector& v, double d )
   {
      return fourvector( v.x * d, v.y * d, v.z * d, v. w );
   }

   inline void operator *= ( fourvector& v, double d )
   {
      v.x *= d;
      v.y *= d;
      v.z *= d;
   }

   inline fourvector operator / ( const fourvector& v, double d )
   {
      if( d <= 0.0 )
         throw std::runtime_error( "cannot make w-component <= 0" );
 
      return { v.x, v.y, v.z, v.w * d };
   }

   inline void operator /= ( fourvector& v, double d )
   {
      if( d <= 0.0 )
         throw std::runtime_error( "cannot make w-component <= 0" ); 

      v.w *= d; 
   }

   inline std::ostream& operator << ( std::ostream& stream, 
                                      const fourvector& v )
   {
      stream << "[ " << v.x << ", " << v.y << ", " << v.z << "; ";
      stream << v.w << " ]";
      return stream;
   }

}


#endif


