
#ifndef LINALG_GENERALMATRIX_INCLUDED
#define LINALG_GENERALMATRIX_INCLUDED  1

#include <vector> 
#include "issingular.h"
#include "generalvector.h"

namespace linalg
{

   template< typename X >
   struct generalmatrix
   {
      size_t dim_i;
      size_t dim_j;

      std::vector<X> repr;   // Has size dim_i * dim_j.

      generalmatrix( ) = delete;

   public:     
      explicit generalmatrix( size_t dim_i, size_t dim_j,
                              const X& x = normdef<X> { }. zero( ))
         : dim_i{ dim_i },
           dim_j{ dim_j },
           repr{ std::vector<X> ( dim_i * dim_j, x ) }
      { }

      generalmatrix( std::initializer_list< std::initializer_list<X>> init );  
         // The matrix, listed as column.
 
      typename std::vector<X> :: iterator operator [] ( size_t i ) 
         { return repr. begin( ) + dim_j * i; } 
      typename std::vector<X> :: const_iterator operator [] ( size_t i ) const
         { return repr. cbegin( ) + dim_j * i; } 
      
      size_t I_dimension( ) const { return dim_i; }
      size_t J_dimension( ) const { return dim_j; }

      generalmatrix<X> operator - ( ) const
      {
         auto res = *this;
         for( X& x : res. repr ) x = -x;
         return res;
      }

      template< typename D > generalmatrix& leftmultiply( D d )
      {  
         for( X& x : repr ) x = d * x; return *this;
      }

      template< typename D > generalmatrix& rightmultiply( D d ) 
      {
         for( X& x : repr ) x = x * d; return *this;
      }

      template< typename D > generalmatrix& operator *= ( D d )
      {
         for( X& x : repr ) x *= d; return *this;
      }

      template< typename D > generalmatrix& operator /= ( D d )
      {
         for( X& x : repr ) x /= d; return *this;
      } 
      
      void print( std::ostream& out ) const
      {
         for( size_t i = 0; i < dim_i; ++ i )
         {
            if( i == 0 ) out << '/';
            else
            {
               if( i + 1 < dim_i ) out << '|';
               else out << '\\';
            }
     
            for( size_t j = 0; j < dim_j; ++ j )
               out << "      " << (*this)[i][j];
           
            out << "    "; 
            if( i == 0 ) out << '\\';
            else
            {
               if( i + 1 < dim_i ) out << '|';
               else out << '/';
            }
            out << '\n';
         } 
      } 


      generalmatrix<X> inverse( double epsilon, X one = 1 ) const; 
         // This method throws the dreaded issingular exception.
         // one is the one of type X. It is used for filling the initial
         // matrix with 1-s on the diagonal. 
         // epsilon is the norm below which an element will be treated
         // as zero. Note that norm has quadratic nature.
         // If you are certain that the matrix is non-singular, you
         // can use epsilon = 0. 

      // Methods that are used for sweeping:

      void exchange_rows( size_t i1, size_t i2 );
      void multiply_row( size_t i, X x );
      void subtract_rows( size_t i1, X x, size_t i2 );
         // row [i1] -= x * row[ i2 ]; 

   };               


   template< typename X >
   generalmatrix<X> :: generalmatrix( 
             std::initializer_list< std::initializer_list<X>> init )
      : dim_j{ init. size( ) }
   {
      if( dim_j == 0 )
         throw std::runtime_error( "cannot construct matrix with width 0" );
            // It seems that this cannot happen, because C++ will use
            // the default constructor, which is deleted. 

      auto p = init. begin( );
      dim_i = p -> size( );

      if( dim_i == 0 )
         throw std::runtime_error( "cannot construct matrix with height 0" ); 

      repr = std::vector<X> ( dim_i * dim_j, normdef<X> { }. zero( ));
         // There is no feasible way to avoid this.

      size_t j = 0;
      while( p != init. end( ))
      {
         if( p -> size( ) != dim_i )
            throw std::runtime_error("columns in matrix must have equal size");
         size_t i = 0;
         for( auto q = p -> begin( ); q != p -> end( ); ++ q )
            (*this)[ i++ ][j] = *q;

         ++ p; ++ j;
      }
   }


   // There are quite a few multiplication operators: 

   // Multiplication with an element:

   template< typename X, typename D >
   inline generalmatrix<X> operator * ( generalmatrix<X> m, D d )
   {
      m. rightmultiply(d); return m;
   }

   template< typename X, typename D >
   inline generalmatrix<X> operator * ( D d, generalmatrix<X> m )
   {
      m. leftmultiply(d); return m;
   }

   template< typename X, typename D >
   inline generalmatrix<X> operator / ( generalmatrix<X> m, D d )
   {
      m /= d; return m;
   }


   // Multiplication with a general vector:

   template< typename X >
   generalvector<X> operator * ( const generalmatrix<X> & m,
                                 const generalvector<X> & v )
   {
      if( m. J_dimension( ) != v. dimension( ))
         throw std::runtime_error( 
            "matrix * vector: dimension of vector does not fit" );

      generalvector<X> res = generalvector<X> ( m. I_dimension( ));
      for( size_t i = 0; i < m. I_dimension( ); ++ i )
      {
         for( size_t j = 0; j < m. J_dimension( ); ++ j )
            res [i] += m[i][j] * v[j];
      }
      return res; 
   }

   template< typename X >
   generalvector<X> operator * ( const generalvector<X> & v,
                                 const generalmatrix<X> & m )
   {
      if( m. I_dimension( ) != v. dimension( ))
         throw std::runtime_error( 
            "vector * matrix: dimension of vector does not fit" );

      generalvector<X> res = generalvector<X> ( m. J_dimension( ));
      for( size_t j = 0; j < m. J_dimension( ); ++ j )
      {
         for( size_t i = 0; i < m. I_dimension( ); ++ i )
            res [j] += v[i] * m[i][j];
      }
      return res;
   }

   template< typename X >
   generalmatrix<X> operator * ( const generalmatrix<X> & m1,
                                 const generalmatrix<X> & m2 )
   {
      if( m1. J_dimension( ) != m2. I_dimension( ))
         throw std::runtime_error( 
            "matrix * matrix: width of first differs from height of second" );

      generalmatrix<X> res = 
            generalmatrix<X> ( m1. I_dimension( ), m2. J_dimension( ));

      for( size_t i = 0; i < m1. I_dimension( ); ++ i )
         for( size_t j = 0; j < m2. J_dimension( ); ++ j )
         {
            for( size_t k = 0; k < m1. J_dimension( ); ++ k )
               res [i][j] += m1 [i][k] * m2 [k][j];
         }

      return res;
   }

   template< typename X > 
   inline std::ostream& operator << ( std::ostream& out, 
                                      const generalmatrix<X> & m )
   {
      m. print( out );
      return out;
   }


   ///////////////////////////////////////////////////////////
   //      Methods that are needed for sweeping:            //
   //      All multiplications by an X must be front.       //
   ///////////////////////////////////////////////////////////   


   template< typename X >
   inline void
   generalmatrix<X> :: exchange_rows( size_t i1, size_t i2 )
   {
      auto p1 = (*this)[i1];
      auto p2 = (*this)[i2];

      for( size_t j = 0; j < J_dimension( ); ++ j )
         std::swap( *p1 ++, *p2 ++ );
   }

   template< typename X >
   inline void 
   generalmatrix<X> :: multiply_row( size_t i, X x )
   {
      auto p = (*this)[i];
      for( size_t j = 0; j < J_dimension( ); ++ j )
      {
         (*p) = x * (*p);
         ++ p;
      }
   }

   template< typename X > 
   inline void 
   generalmatrix<X> :: subtract_rows( size_t i1, X x, size_t i2 )
   {
      auto p1 = (*this)[i1];
      auto p2 = (*this)[i2];
      for( size_t j = 0; j < J_dimension( ); ++ j )
      {
         *p2 = *p2 - x * (*p1);
         ++ p1; ++ p2;
      }
   }

   template< typename X >
   generalmatrix<X> generalmatrix<X> :: inverse( double epsilon,
                                        X one ) const
   {
      if( dim_i != dim_j )
         throw std::runtime_error( "inverse: matrix is not square" );

      if( epsilon < 0 )
         throw std::runtime_error( "epsilon must be >= 0" );

      auto m = *this;
      generalmatrix<X> inv = generalmatrix<X> ( dim_i, dim_i ); 
      for( size_t i = 0; i < dim_i; ++ i )
         inv[i][i] = one;

      normdef<X> def; 

      size_t failures = 0; 
      for( size_t j = 0; j < dim_j; ++ j )
      {

#if 0
         std::cout << "----------------------------------\n";
         std::cout << "sweeping column " << j << ":\n";
         std::cout << m << "\n";
         std::cout << inv << "\n";
#endif

         double maxnorm = epsilon;
         size_t maxi = 0;
         for( size_t i = j; i < dim_i; ++ i )
         {
            double n = def( m[i][j] );
            if( n > maxnorm )
            {
               maxnorm = n;
               maxi = i; 
            }
         }

         if( maxnorm == epsilon )
            failures = failures + 1;
         else
         {
            // We exchange the best row with position j:

            if( maxi != j )
            {
               m. exchange_rows( j, maxi );
               inv. exchange_rows( j, maxi );
            }

            // Using j, we clean the other rows:

            for( size_t i = 0; i < dim_i; ++ i ) 
               if( i != j )
               {
                  X pivot = m[i][j] / m[j][j];
                  m. subtract_rows( j, pivot, i );
                  m[i][j] = def. zero( );              // Looks better. 
                  inv. subtract_rows( j, pivot, i ); 
               }
         }
      }

      if( failures ) throw issingular( dim_i - failures, dim_i );

      for( size_t i = 0; i < dim_i; ++ i )
      {
         inv. multiply_row( i, 1.0 / m[i][i] );
      }

      return inv;
   }

}

#endif

