
// A rototrans is a special case of an affinemap
//    F(x) = m.x + v, where m is a rotation.
// 
// rototrans is preferable over affinemap, because it can
// be more easily composed, and inverted. We represent m by a quaternion.
// rototrans should be used for representing the position and orientation
// of rigid objects.
//
// Hans de Nivelle, October 2014.

#ifndef LINALG_ROTOTRANS_INCLUDED
#define LINALG_ROTOTRANS_INCLUDED  1

#include "vector.h"
#include "quaternion.h"
#include "affinemap.h"

#include <iostream>

namespace linalg
{

   struct rototrans 
   {
      quaternion q;    
      vector x0;

      // Default is identity: 

      rototrans( )
         : q{1,0,0,0},
           x0{0,0,0} 
      { } 

      rototrans( quaternion q, vector x0 )
         : q{q}, x0{x0} 
      { } 

      vector apply( vector v ) const
         { return x0 + q. rotate(v); } 

      vector operator( ) ( vector v ) const
         { return x0 + q. rotate(v); }

      vector apply_inv( vector v ) const
         { return q. conjugate( ). rotate( v - x0 ); }

      vector only_rotate( vector v ) const
         { return q. rotate(v); }
          
      rototrans inverse( ) const 
         { return { q. conjugate( ), - q. conjugate( ). rotate(x0) }; }


      // Create a rototrans with rotation q, and which maps p to 
      // the required image of p. 
   
      static rototrans require( quaternion q,
                                vector p, vector image_of_p )
      {
         return { q, image_of_p - q. rotate(p) };
      }

      rototrans& composewith( const rototrans& rt );
         // Left compose with rt. 
         //    (*this) := rt * (*this). 

      affinemap getaffinemap( ) const
         { return { q. getrotationmatrix( ), x0 }; } 
   };


   // ( t1 * t2 ) means: First t2, then t1.
   // As a consequence, we have (t1*t2)(v) = t1(t2(v)).

   inline rototrans operator * ( const rototrans& rt1, const rototrans& rt2 )
   {
      return { rt1.q * rt2.q,  rt1.x0 + rt1.q. rotate( rt2.x0 ) }; 
   }


   inline rototrans& rototrans::composewith( const rototrans& rt )
   {
      x0 = rt.q.rotate(x0) + rt.x0; q = rt.q * q; return *this;      
   }


   inline 
   std::ostream& operator << ( std::ostream& out, const rototrans& rt )
   {
      out << "rototrans( q = " << rt.q << ", x0 = " << rt.x0 << " )";
      return out;
   }

}

#endif


