
// (Re)Written by Hans de Nivelle, November 2016.

#ifndef LINALG_SPINSPEED_INCLUDED
#define LINALG_SPINSPEED_INCLUDED  1

// Spinspeed is the derivative of a rototrans over time.
// It can represented by an angular velocity omega, and
// the speed v0 at (0,0,0). For every point x, the spinspeed maps
// x to omega * x + v0. 
//
// Be careful with rotations around an axis that is far away from
// the origin. v0 can be very high, and this will result in loss
// of floating point accurracy. 


#include "vector.h"

namespace linalg
{

   struct spinspeed
   {   
      vector omega;       // Angular velocity. 
      vector v0;          // Speed at zero.

      spinspeed( ) 
         : omega{0,0,0},
           v0{0,0,0}
      { }

      spinspeed( const vector& omega, const vector& v0 )
         : omega{ omega },
           v0{ v0 }
      {
      }

      vector apply( vector x ) const
         { return crossproduct( omega, x ) + v0; }

      vector operator( ) ( vector x ) const 
         { return crossproduct( omega, x ) + v0; }


      // Spinspeed with angular component omega, and which has 
      //    S(x) = vx.

      static spinspeed
      require( const vector& omega, const vector& x, const vector& vx )
      {
         return { omega, vx - crossproduct( omega, x ) };
      }

   };

   // For additive operators, we have   (V1+V2)(v) = V1(v) + V2(v).
   // (I am not sure if these are meaningful operators at all. Maybe
   //  they should not exist. By 'not meaningful' I mean: Not typeable 
   //  in a suitably chosen type system.)
 
   inline spinspeed operator + ( const spinspeed& sp1, const spinspeed& sp2 )
   {
      return { sp1. omega + sp2. omega, sp1. v0 + sp2. v0 }; 
   }

   inline spinspeed& operator += ( spinspeed& sp1, const spinspeed& sp2 )
   {
      sp1. omega += sp2. omega;
      sp1. v0 += sp2. v0; 
      return sp1;
   }

   inline spinspeed operator - ( const spinspeed& sp1, const spinspeed& sp2 )
   {
      return { sp1. omega - sp2. omega, sp1. v0 - sp2. v0 };
   }

   inline spinspeed& operator -= ( spinspeed& sp1, const spinspeed& sp2 )
   {
      sp1. omega -= sp2. omega;
      sp1. v0 -= sp2. v0;
      return sp1;
   }

   // For multiplicative operators, we have (V*d)(v) = d * ( V(d) ):
   // They might me meaningful.

   inline spinspeed operator - ( const spinspeed& sp )
   {
      return { - sp. omega, - sp. v0 };
   }

   inline spinspeed operator * ( const spinspeed& sp, double d )
   {
      return { sp. omega * d, sp. v0 * d };
   }

   inline spinspeed operator * ( double d, const spinspeed& sp )
   {
      return { d * sp. omega, d * sp. v0 }; 
   }

   inline spinspeed& operator *= ( spinspeed& sp, double d )
   {
      sp. omega *= d;
      sp. v0 *= d;
      return sp; 
   }

   inline spinspeed operator / ( const spinspeed& sp, double d )
   {
      return { sp. omega / d, sp. v0 / d };
   }

   inline spinspeed& operator /= ( spinspeed& sp, double d )
   {
      sp. omega /= d;
      sp. v0 /= d;
      return sp;
   }

   inline std::ostream& operator << ( std::ostream& out, const spinspeed& sp )
   {
      out << "spinspeed( ";
      out << "omega = " << sp. omega << ", ";
      out << "v0 = " << sp. v0 << " )";
      return out;
   }

}

#endif

