
// Written by Hans de Nivelle, 10 sept 2010.
// Updated 5 oct 2014.

#ifndef MATH_COMPLEX_INCLUDED
#define MATH_COMPLEX_INCLUDED 1

#include <iostream>
#include <math.h>
#include <list> 

namespace math
{

   struct complex
   {
      double re;
      double im;
    
      complex( )
         : re{ 0.0 },
           im{ 0.0 }
      { }

      complex( double r )
         : re{r},
           im{0.0}
      { } 

      constexpr complex( double re, double im )
         : re{ re },
           im{ im }
      { }

      double norm( ) const
      {
         return re*re + im*im;
      }

      double length( ) const
      {
         return sqrt( norm( ));
      }

      complex conjugate( ) const
      {
         return complex( re, -im );
      }

      complex inverse( ) const
      {
         double n = norm( );
         return complex( re / n, -im / n );
      }
   };


   inline complex operator + ( complex c1, complex c2 )
   {
      return { c1. re + c2. re, c1. im + c2. im };
   }

   inline complex operator - ( complex c1, complex c2 )
   {
      return { c1. re - c2. re, c1. im - c2. im };
   }

   inline complex operator - ( complex c )
   {
      return { -c. re, -c. im };
   }

   inline complex& operator += ( complex& c1, complex c2 )
   {
      c1. re += c2. re;
      c1. im += c2. im;
      return c1;
   }

   inline complex& operator -= ( complex& c1, complex c2 )
   {
      c1. re -= c2. re;
      c1. im -= c2. im;
      return c1; 
   }

   inline complex operator * ( complex c, double d )
   {
      return { c. re * d, c. im * d };
   }

   inline complex operator * ( double d, complex c )
   {
      return { d * c. re, d * c. im };
   }

   inline complex operator * ( complex c1, complex c2 )
   {
      return { c1. re * c2. re - c1. im * c2. im,
               c1. re * c2. im + c1. im * c2. re };
   }

   inline complex& operator *= ( complex& c, double d )
   {
      c. re *= d;
      c. im *= d;
      return c;
   }

   inline complex& operator *= ( complex& c1, const complex& c2 )
   {
      c1 = c1 * c2;  
      return c1; 
   }

   inline complex operator / ( complex c, double d )
   {
      return { c. re / d, c. im / d };
   }

   inline complex& operator /= ( complex& c, double d )
   {
      c. re /= d;
      c. im /= d;
      return c;
   }

   inline complex operator / ( complex c1, const complex& c2 )
   {
      return c1 * c2. conjugate( ) / c2. norm( );
   }

   inline complex& operator /= ( complex& c1, const complex& c2 )
   {
      c1 *= c2. conjugate( ) / c2. norm( );
      return c1;
   }

   // Allows to write 4.0_i:

   constexpr complex operator "" _i( long double x ) 
   {
      return complex( 0, x ); 
   }

   std::ostream& operator << ( std::ostream& out, complex c );

   inline complex exp( complex c )
      { return ::exp( c.re ) * complex( ::cos( c.im ), ::sin( c.im )); }

   inline complex sin( complex x ) 
      { return ( exp( 1.0_i * x ) - exp( -1.0_i * x )) / ( 2.0_i ); } 

   inline complex cos( complex x )    
      { return ( exp( 1.0_i * x ) + exp( -1.0_i * x ) ) / 2.0; } 

}

#endif


