
// A few integration methods.
// VERY MUCH UNFINISHED.

#ifndef MATH_INTEGRATION_INCLUDED
#define MATH_INTEGRATION_INCLUDED 1

namespace math
{

   // The type I must have a default constructor and
   // a membor I::vector that will be the type on which we are integrating.
   // Const elements i of I must be applicable on
   // elements of I::vector. The result must be of I::vector again.
   // For example:
   // 
   //    struct integrator
   //    {
   //       typedef statevector vector;
   //       vector operator( ) ( const vector& x ) const { return F(x); }
   //    };
   //
   // In the code, define I i, and call 
   // runge_kutta( i, x, h ); 

   template< typename I >
   typename I::vector
   runge_kutta1( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      return x + h * k1;
   }

   // We have two methods of order two. The first method is Heun's method.
   // runge_kutta22 is approximately two times better than 
   // runge_kutta21. 

   template< typename I >
   typename I::vector
   runge_kutta21( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + h * k1 );

      return x + h * ( 0.5 * k1 + 0.5 * k2 );
   }

   template< typename I >
   typename I::vector
   runge_kutta22( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + 0.5 * h * k1 );

      return x + h * k2; 
   }


   // A third order method:

   template< typename I >
   typename I::vector
   runge_kutta31( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + h * (2.0/3.0) * k1 );
      auto k3 = i( x + h * ( (1.0/3.0) * k1 + (1.0/3.0) * k2 ) );

      return x + h * ( (1.0/4.0) * k1 + (3.0/4.0) * k3 );
   }


   // The standard Runge Kutta method:

   template< typename I > 
   typename I::vector
   runge_kutta41( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + 0.5 * h * k1 );
      auto k3 = i( x + 0.5 * h * k2 );
      auto k4 = i( x + h * k3 );

      return x + ( 1.0 / 6.0 ) * h * ( k1 + 2.0 * k2 + 2.0 * k3 + k4 );
   }


   // A fourth order method by J.Kuntzmann. Accuracy is 
   // approximately two times better than standard Runge Kutta. 

   template< typename I >
   typename I::vector
   runge_kutta4_kuntzmann( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + h * (2.0/5.0) * k1 );
      auto k3 = i( x + h * ( (-3.0/20.0) * k1 + (3.0/4.0) * k2 ) );
      auto k4 = i( x + h * ( (19.0/44.0) * k1 + (-15.0/44.0) * k2 + 
                             (40.0/44.0) * k3 ) );

      return x + h * ( (55.0/360.0) * k1 + (125.0/360.0) * k2 +
                       (125.0/360.0) * k3 + (55.0/360.0) * k4 );
   }


   // A fifth-order Runge Kutta method with six stages:
   // (Page 92 in J.C.Butcher, Numerical Methods for Ordinary Differential
    // Equations.) 

   template< typename I > 
   typename I::vector
   runge_kutta5( const I& i, const typename I::vector& x, double h )
   {
      auto k1 = i( x );
      auto k2 = i( x + h * (1.0/4.0) * k1 );
      auto k3 = i( x + h * ( (1.0/8.0) * k1 + (1.0/8.0) * k2 ) );
      auto k4 = i( x + h * (1.0/2.0) * k3 );
      auto k5 = i( x + h * ( (3.0/16.0) * k1 + (-3.0/8.0) * k2 +
                             (3.0/8.0) * k3 + (9.0/16.0) * k4 ) );
      auto k6 = i( x + h * ( (-3.0/7.0) * k1 + (8.0/7.0) * k2 +
                             (6.0/7.0) * k3 + (-12.0/7.0) * k4 + 
                             (8.0/7.0) * k5 ) );
      
      return x + h * 
         (   ( 7.0 / 90.0 ) * k1 + ( 32.0 / 90.0 ) * k3 + 
             ( 12.0 / 90.0 ) * k4 + ( 32.0 / 90.0 ) * k5 +
             ( 7.0 / 90.0 ) * k6 );
   }

}

#endif


