
// Written by Hans de Nivelle, May 2012.
// Separated interpolator from spline in October 2014.

#include "interpolator.h"
#include "../linalg/bandmatrix.h"


const math::interpolator::polynomial& 
math::interpolator::findpoly( double x ) const
{

   if( x < table[0]. b ) return table[0];
   if( x >= table[ table. size( ) - 1 ]. b ) 
      return table [ table. size( ) - 1 ];

   unsigned int lower = 0;
   unsigned int supremum = table. size( ) - 1;

   // We do binary search until the range small enough: 

   while( lower + 3 < supremum )
   {
      unsigned int middle = ( lower + supremum ) / 2;
      if( x < table [ middle ]. b )
         supremum = middle;
      else
         lower = middle;
   }

   // After that, we do linear search. A solution is guaranteed
   // to exist. 

   while( table [ lower ]. b <= x ) 
      ++ lower;

   -- lower;  

   return table [ lower ];
}

math::interpolator::interpolator( std::initializer_list< polynomial > table )
{
   for( auto p = table. begin( ); p != table. end( ); ++ p )
      ( this -> table ). push_back( *p );
}

math::interpolator::interpolator( const std::vector< polynomial > & table )
   : table{ table }
{ }

math::interpolator::interpolator( std::vector< polynomial > && table )
   : table{ std::move( table ) }
{ }

std::ostream& math::operator << ( std::ostream& stream,
                                  const interpolator::polynomial& p )
{
   stream << p. b << "  : F(X) =  ";
   stream << p. a0 << " + " << p. a1 << " * X + ";
   stream << p. a2 << " * X^2 + " << p. a3 << " * X^3 ";

   return stream;
}


std::ostream& math::operator << ( std::ostream& stream, const interpolator& i )
{
   stream << "Interpolator:\n";
   for( auto p = i. table. begin( ); p != i. table. end( ); ++ p )
   {
      stream << "   " << *p << "\n";
   }
   return stream;
}


math::interpolator 
math::interpolator::block( const std::map< double, double > & values )
{
   if( values. size( ) == 0 )
      throw std::runtime_error( "cannot create empty interpolator" );

   std::vector< interpolator::polynomial > res;
   
   auto p1 = values. begin( );
   auto p2 = values. begin( ); ++ p2;

   while( p2 != values. end( ))
   {
      res. push_back( 
            interpolator::polynomial( p1 -> first, p1 -> second, 0, 0, 0 )); 
      ++ p1;
      ++ p2;
   }
   res. push_back( 
           interpolator::polynomial( p1 -> first, p1 -> second, 0, 0, 0 ));
   return res;
}

math::interpolator 
math::interpolator::linear( const std::map< double, double > & values )
{
   if( values. size( ) < 2 )
      throw std::runtime_error( 
         "cannot create linear interpolator with less than two points" );

   std::vector< interpolator::polynomial > res;

   auto p1 = values. begin( );
   auto p2 = values. begin( ); ++ p2;

   while( p2 != values. end( ))
   {
      double deriv = 
         (( p2 -> second ) - ( p1 -> second )) /
         (( p2 -> first ) - ( p1 -> first ));

      res. push_back(
           interpolator::polynomial( p1 -> first,
              p1 -> second - deriv * p1 -> first, deriv, 0, 0 ));
      ++ p1;
      ++ p2;
   }
   return res;
}


namespace
{
   // The order of the columns in the matrix is always 
   // [ A00, A01, A02, A03;  A10, A11, A12, A13;  A20, A21, A22, A23; ... ]


   // Write the deriv-th derivative in position x for polynomial p
   // in row row.

   inline void writepolynomial( linalg::bandmatrix<double,5> & mat,
                                unsigned int row, unsigned int pol, 
                                unsigned int deriv, double x,
                                int sign )
   {
      if( row >= mat. dimension( ))
         throw std::runtime_error( "writepolynomial!" );
      if( pol * 4 >= mat. dimension( ))
         throw std::runtime_error( "writepolynomial!" );

      double x2 = x * x;
      double x3 = x2 * x;

      switch( deriv )
      {
      case 0:
         mat [ row ] [ 4 * pol + 0 ] = sign * 1.0;
         mat [ row ] [ 4 * pol + 1 ] = sign * x;
         mat [ row ] [ 4 * pol + 2 ] = sign * x2;
         mat [ row ] [ 4 * pol + 3 ] = sign * x3;
         return;

      case 1:
         mat [ row ] [ 4 * pol + 0 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 1 ] = sign * 1.0;
         mat [ row ] [ 4 * pol + 2 ] = sign * 2.0 * x;
         mat [ row ] [ 4 * pol + 3 ] = sign * 3.0 * x2;
         return;

      case 2:
         mat [ row ] [ 4 * pol + 0 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 1 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 2 ] = sign * 2.0;
         mat [ row ] [ 4 * pol + 3 ] = sign * 6.0 * x;
         return;

      case 3:
         mat [ row ] [ 4 * pol + 0 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 1 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 2 ] = sign * 0.0;
         mat [ row ] [ 4 * pol + 3 ] = sign * 6.0;
         return;

      default:
         throw std::runtime_error( 
            "writepolynomial:cannot differentiate more than 4 times" ); 

      }
   }

   inline void write_value( linalg::bandmatrix<double,5> & mat, 
                            linalg::generalvector<double> & vect,
                            unsigned int row, unsigned int pol, 
                            unsigned int deriv, double x, double y )
                                          
   {
      writepolynomial( mat, row, pol, deriv, x, 1 );
      vect [ row ] = y;
   }


   // Write the fact that the deriv-th derivatives of polynomials
   // pol1 and pol2 are equal in point x.

   inline void write_equal( linalg::bandmatrix<double,5> & mat, 
                            linalg::generalvector<double> & vect,  
                            unsigned int row, 
                            unsigned int pol1, unsigned int pol2, 
                            unsigned int deriv, double x )
   {
      if( pol1 == pol2 )
         throw std::runtime_error( "cannot equate polynomial to itself" );
      writepolynomial( mat, row, pol1, deriv, x, 1 );
      writepolynomial( mat, row, pol2, deriv, x, -1 );
      vect [ row ] = 0.0;
   }

}


math::interpolator 
math::interpolator::spline( const std::map< double, double > & values, 
                            double deriv2begin, double deriv2end )
{
   if( values. size( ) < 3 )
      throw std::runtime_error( "spline: need at least 3 values" );

   linalg::bandmatrix<double,5> mat( 4 * values. size( ) - 4 ); 
   linalg::generalvector<double> vect( 4 * values. size( ) - 4 );

   auto p1 = values. begin( );
   auto p2 = values. begin( ); ++ p2;
      // p2 will always be one step ahead of p1.

   unsigned int row = 0; 
      // row in which we will write the next equation.

   for( unsigned int j = 0; j + 1 < values. size( ); ++ j )
   {

      // If we are at the beginning, then we write the condition
      // that 2nd derivative must be deriv2begin.

      if( j == 0 )
      {
         write_value( mat, vect, row, j, 2, p1 -> first, deriv2begin );
         row = row + 1;
      }

      // We always write the values at the beginning and at the end 
      // of the interval. 

      write_value( mat, vect, row, j, 0, p1 -> first, p1 -> second );
      row = row + 1;

      write_value( mat, vect, row, j, 0, p2 -> first, p2 -> second );
      row = row + 1;   

      // If we are at the end, we write the condition that 
      // 2nd derivative must be deriv2end. 
      // If not, we write the condition that 1st and 2nd derivative 
      // at the end of the interval must be equal to the 1st and 2nd
      // derivative at the beginning of the next interval.

      if( j + 2 == values. size( ))
      {
         write_value( mat, vect, row, j, 2, p2 -> first, deriv2end );
         row = row + 1;
      }
      else  
      {
         write_equal( mat, vect, row, j, j + 1, 1, p2 -> first );
         row = row + 1;

         write_equal( mat, vect, row, j, j + 1, 2, p2 -> first ); 
         row = row + 1;
      }

      ++ p1; ++ p2;
   }

   try
   {
      linalg::generalvector< double > 
         coeff = mat. apply_inverse( vect, vect. norm( ) * 1.0E-15 );

      // If we are still alive, we have coeff and we can fill in the tables:

      auto p1 = values. begin( );
      auto p2 = values. begin( ); ++ p2;

      std::vector< interpolator::polynomial > table;

      unsigned int i = 0;
      while( p2 != values. end( ))
      {
         table. push_back( 
            interpolator::polynomial( 
                        p1 -> first,
                        coeff[i], coeff[i+1], coeff[i+2], coeff[i+3] ));
         ++ p1;
         ++ p2;
         i += 4;
      }

      return table;
   }
   catch( linalg::issingular s )
   {
      // We have really no idea what to do. 

      std::cout << "------------------------------------\n";
      std::cout << mat << "\n";
      std::cout << vect << "\n";

      std::cout << "cannot construct spline: " << s << "\n";

      throw std::runtime_error( 
         "cannot compute spline. This should not happen" ); 
   }

}


math::interpolator 
math::interpolator::choose( interpolator_name choice,
                            const std::map< double, double > & values,
                            double deriv2begin, double deriv2end )
{
   switch( choice )
   {
   case interpolator_name::block:
      return block( values );
   case interpolator_name::linear:
      return linear( values );
   case interpolator_name::spline:
      return spline( values, deriv2begin, deriv2end );
   default:
      throw std::runtime_error( "interpolator::choose : unknown choice" ); 
   }
}


