
#ifndef MATH_INTERPOLATOR_INCLUDED
#define MATH_INTERPOLATOR_INCLUDED    1

// This was made on 24.10.2014. 

#include <iostream>
#include <vector>
#include <map>

namespace math
{

   enum class interpolator_name { block, linear, spline };

   // An interpolator consists of a sequence of 3d degree polynomials,
   // combined with a lower bound b from which the polynomial
   // can be applied. Polynomial P[i] will be applied on x if x >= P[i].b,
   // and x < P[i+1].b. 

   struct interpolator
   {
      // A polynomial consists of a number b and the polynomial itself. 
      // The polynomial can be applied on x when x >= b.

      struct polynomial
      {
         double b;

         double a0;
         double a1; 
         double a2;
         double a3;
            // The meaning is clear:
            // If b <= x, then the polynomial 
            //    a0 + a1.x + a2.x^2 + a3.x^3 can be applied. 

         polynomial( double b, double a0, double a1, double a2, double a3 )
            : b{b}, 
              a0{a0}, a1{a1}, a2{a2}, a3{a3}
         { }

         double apply( double x1 ) const
         {
            double x2 = x1 * x1;
            double x3 = x2 * x1;
            return a0 + a1 * x1 + a2 * x2 + a3 * x3;
         }

         double derivative( double x1 ) const
         {
            double x2 = x1 * x1;
            return a1 + 2.0 * a2 * x1 + 3.0 * a3 * x2;
         }

         double derivative2( double x1 ) const
         {
            return 2.0 * a2 + 6.0 * a3 * x1;
         }

         polynomial derivative( ) const
         { 
            return polynomial( b, a1, 2.0 * a2, 3.0 * a3, 0.0 );
         }
      };


      std::vector< polynomial > table;

   public: 
      // Polynomials must be sorted by b:

      interpolator( ) = delete; 

      // For each of the constructors, the table must be sorted.

      interpolator( std::initializer_list< polynomial > table );
      interpolator( const std::vector< polynomial > & table );
      interpolator( std::vector< polynomial > && table );

      const polynomial& findpoly( double x ) const;
         // Find the polynomial that will be applied on x. This is table[i] 
         // with table[i].b <= x, for the highest possible i. 
 
      double apply( double x ) const 
         { return findpoly(x). apply(x); } 

      double derivative( double x ) const 
         { return findpoly(x). derivative(x); }

      double derivative2( double x ) const 
         { return findpoly(x). derivative2(x); }

         // Returns the value, or the 1st or 2nd derivative, in x.  
         // These three functions also return a value when
         // x is not in the domain. This value is not necessarily meaningful
         // when x is far away from the domain.  

      double operator( ) ( double x ) const { return apply(x); } 

      friend std::ostream& operator << ( std::ostream& , const interpolator& ); 
      friend std::ostream& operator << ( std::ostream& , 
                                         const interpolator::polynomial& ); 

      static interpolator block( const std::map< double, double > & values );
         // Construct a piecewise constant interpolator. 
         // This is the simplest possible interpolator.

      static interpolator linear( const std::map< double, double > & values );
         // Construct a piecewise linear interpolator.

      static interpolator spline( const std::map< double, double > & values,
                                  double deriv2begin, double deriv2end );
         // Construct a spline interpolator. 

      static interpolator choose( interpolator_name choice,
                                  const std::map< double, double > & values,
                                  double deriv2begin, double deriv2end ); 

   };

   std::ostream& operator << ( std::ostream& , const interpolator& ); 
   std::ostream& operator << ( std::ostream& , 
                               const interpolator::polynomial& ); 

}

#endif

