
#include "rigid.h"

// Written by Hans de Nivelle, August 2010.
//                             Revised 07.11.2012. 
//                             And revised again in January 2017. 


linalg::matrix 
physics::pointmass::inertia_matrix( const linalg::vector ref ) const 
{
   const linalg::vector p = position - ref; 

   double yyzz = p.y * p.y + p.z * p.z; 
   double xxzz = p.x * p.x + p.z * p.z; 
   double xxyy = p.x * p.x + p.y * p.y;

   double xy = - p.x * p.y;
   double xz = - p.x * p.z; 
   double yz = - p.y * p.z;

   return mass * linalg::matrix( { yyzz, xy, xz },
                                 { xy, xxzz, yz },
                                 { xz, yz, xxyy } ); 
}


physics::mass_distribution::mass_distribution(
           const std::vector< pointmass > & distr )
      : total{0},
        center{0,0,0},
        inertia{ linalg::matrix::zero( ) }
{
   if( distr. size( ) == 0 )
      throw std::runtime_error( "cannot compute mass center "\
                                   "from empty vector of point masses" );

   for( const auto& m : distr )
   {
      if( m. mass < 0.0 )
         throw std::runtime_error( "negative mass makes no sense" );

      total += m. mass;
      center += m. mass * m. position;
   }
   center /= total;

   for( const auto& m : distr )
      inertia += m. inertia_matrix( center );
}


std::ostream& 
physics::operator << ( std::ostream& out, const mass_distribution& m )
{
   out << "Mass distribution:\n";
   out << "   total =    " << m. total << ", ";
   out << "   center =   " << m. center << "\n";
   out << "   inertia = \n" << m. inertia << "\n";
   return out;
}


physics::rigid::rigid( const linalg::rototrans& xxxx,
                       const linalg::spinspeed& vvvv,
                       const mass_distribution& distr ) 
   : state{ xxxx, vvvv },
     distr{ distr },
     inertia_inv{ distr. inertia. inverse( 1.0E-12 ) }  
{
   // Now we have xxxx,vvvv based on internal coordinates.
   // They must be based on internal coordinates, using 
   // mass center. 

   ( this -> xxxx ).x0 += xxxx. only_rotate( distr. center );
   ( this -> vvvv ).v0 += 
      linalg::crossproduct( vvvv.omega, xxxx. only_rotate( distr. center ));
}


physics::force_torque
physics::rigid::effect_internal( const linalg::vector& f,
                                 const linalg::vector& acting_point ) const
{
   force_torque res_int = 
      { f, linalg::crossproduct( acting_point - distr. center, f ) };
   return
      { xxxx.q. rotate( res_int. f ), xxxx.q. rotate( res_int. tq ) };
}


physics::force_torque
physics::rigid::effect_external( const linalg::vector& f,
                                 const linalg::vector& acting_point ) const
{
   linalg::vector p = acting_point - xxxx.x0;
   return { f, linalg::crossproduct( p, f ) };
}


linalg::rototrans
physics::rigid::int2ext( ) const
{
   return linalg::rototrans::require( xxxx.q, distr. center, xxxx.x0 );
}

linalg::rototrans
physics::rigid::ext2int( ) const
{
   return linalg::rototrans::require( xxxx.q. inverse( ), 
                                      xxxx.x0, distr. center );
}

linalg::spinspeed physics::rigid::intspeed( ) const
{
   return linalg::spinspeed::require( xxxx.q. inverse( ). rotate( vvvv. omega ),
                                      distr. center,
                                      xxxx.q. inverse( ). rotate( vvvv.v0 ));
}

physics::state_acc 
physics::rigid::accelerations( const force_torque& ft ) const
{

   // We need torque and angular velocity in internal coordinates:

   linalg::vector tq_internal = xxxx.q. inverse( ). rotate( ft.tq );
   linalg::vector omega_internal = xxxx.q. inverse( ). rotate( vvvv. omega );
   
   // We can now compute angular accelaration in internal coordinates. 
   // We use the equation
   //    torque = M(alpha) + omega X M( omega ).

   linalg::vector alpha_internal = 
      inertia_inv(  
          tq_internal -
          linalg::crossproduct( omega_internal, 
                                distr. inertia( omega_internal )));

   return { ft.f / distr. total, xxxx.q.rotate( alpha_internal ) };
}


std::ostream& physics::operator << ( std::ostream& stream, const rigid& rig )  
{
   stream << "Rigid Object:\n";
   stream << rig. distr << "\n";
   stream << rig. inertia_inv << "\n";
   const state& s = rig;
   stream << s;
   stream << "   (of mass center)\n";
   return stream;
}


