
// This is only a first attempt. Should be rewritten to work with
// Runge-Kutta methods.

#ifndef TWODIMENSIONAL_INCLUDED
#define TWODIMENSIONAL_INCLUDED 1

#include <string>
#include <iostream>

#include "linalg/vector.h"
#include "airfoil_mach0.h"

struct force_effect
{
   linalg::vector linear; 
   double torque; 

   force_effect( ) :
      linear{0,0,0},
      torque{0}
   { }

   
   force_effect( const linalg::vector& lin, double tq )
      : linear{ lin },
        torque{ tq }
   { }
      // Unfortunately, this wouldn't work in 3 dimensions. 

   force_effect( const linalg::vector& lin, 
                 const linalg::vector& acting_point )
      : linear{ lin },
        torque{ acting_point.x * lin.y -
                acting_point.y * lin.x }
   { }

};

inline force_effect operator + ( force_effect e1, force_effect e2 )
{
   return { e1. linear + e2. linear, e1. torque + e2. torque };
}



struct twodimensional
{
   std::string name;  // Whatever you want to put here.
   
   double mass;       // Total mass of aircraft. 
   double inertia;    // Inertia moment of aircraft. 

   double rho;        // Air density.

   linalg::vector position;   // In LRU. The z-coordinate is not used. 
   double theta;              // Pitch angle.

   linalg::vector v0;     // V0 of rigid speed function. Only x,y are used.
                          // This is speed at mass center. Read the slides!
   double omega;          // omega of rigid speed function. 

   // Fuselage is a complicated thing. Firstly, it can produce
   // lift. This lift will have a pressure center,
   // which very probably will contribute to instability.
   // Secondly, I cannot find reasonable data.
   // Thirdly, the fuselage can have different sizes in three dimensions.
   // (Only two of them are relevant here.)

   airfoil_mach0 fuselage_coefficients;
   linalg::vector fuselage_position;      
      // Position of nose. Acting point of lift and drag will be
      // computed through chord and pitching moment. Z-coordinate
      // is not used. 

   linalg::vector fuselage_surfaces;
      // X: front facing surface.
      // Y: up/down facing surface. It is crazy to put this in the 
      //    Y-coordinate, but it is a consequence of the fact that
      //    we are two dimensional.
      // Z: Not used. 

   double fuselage_length;
 
   airfoil_mach0 wing_coefficients; 
   linalg::vector wing_position; 
                         // Position of leading edge of wing.
   double wing_theta;    // a.o.a., with which the wings are
                         // connected to the air frame.
   double wing_surface;  // Surface area of wing.
   double wing_chord;    // Chord length of wing.


   linalg::vector engine_position; 
      // Only X and Y are used.

   double throttle;      // A number between -0.3 and 1.1. 
   double engine_power;  // Power converges to throttle, but may be
                         // different.
   
   airfoil_mach0 elevator_coefficients0;     // Default position. 
   airfoil_mach0 elevator_coefficients1;     // Extended position.
   double elevator_deflection;         // Should be between 0 and 1.
             // Positive has the same effect as increasing elevator_theta,
             // which results in pitch down. 

   linalg::vector elevator_position;
      // Position of leading edge axis of elevators. 
      // (There is a small problem here that we ignore.) 
   double elevator_theta;  // a.o.a, with which the elevators are
                           // connected to the fuselage.  
   double elevator_surface;
   double elevator_chord;

   twodimensional( const std::string& name,
                   double mass, double inertia,
                   double rho, 
                   linalg::vector position, double theta,
                   linalg::vector v0, double omega,
                   const airfoil_mach0& fuselage_coefficients,
                   linalg::vector fuselage_position, 
                   linalg::vector fuselage_surfaces, 
                   double fuselage_length,  
                   const airfoil_mach0& wing_coefficients,
                   linalg::vector wing_position,
                   double wing_theta, 
                   double wing_surface, double wing_chord,
                   linalg::vector engine_position,
                   double throttle, double engine_power,
                   const airfoil_mach0& elevator_coefficients0,
                   const airfoil_mach0& elevator_coefficients1, 
                   double elevator_deflection, 
                   linalg::vector elevator_position,
                   double elevator_theta, 
                   double elevator_surface, double elevator_chord )
      : name{ name },
        mass{ mass }, inertia{ inertia },
        rho{ rho },
        position{ position }, theta{ theta },
        v0{ v0 }, omega{ omega },
        fuselage_coefficients{ fuselage_coefficients },
        fuselage_position{ fuselage_position },
        fuselage_surfaces{ fuselage_surfaces }, 
        wing_coefficients{ wing_coefficients },
        fuselage_length{ fuselage_length }, 
        wing_position{ wing_position },
        wing_theta{ wing_theta },
        wing_surface{ wing_surface }, wing_chord{ wing_chord },
        engine_position{ engine_position },
        throttle{ throttle }, engine_power{ engine_power },
        elevator_coefficients0{ elevator_coefficients0 },
        elevator_coefficients1{ elevator_coefficients1 }, 
        elevator_deflection{ elevator_deflection }, 
        elevator_position{ elevator_position },
        elevator_theta{ elevator_theta }, 
        elevator_surface{ elevator_surface }, 
        elevator_chord{ elevator_chord }
   { }


   // All force_effects are with force in LRU (world coordinates),
   // and torque relative to center of mass. 

   force_effect gravity( double g = 9.80665 ) const
   {
      return force_effect( { 0, -mass * g, 0 }, { 0,0,0 } );
   }

   force_effect fuselage( ) const; 
   force_effect main_wings( ) const; 

   force_effect engines( ) const
   {
      return force_effect( 
                 linalg::vector( cos(theta),sin(theta),0 ) * engine_power *
                           2.0 * 89000.0,
                           linalg::vector( 2 * sin(theta), 
                                          -2 * cos(theta) ,0 ));
   }

   force_effect elevators( ) const;    

};


std::ostream& operator << ( std::ostream& out, const twodimensional& t ); 
std::ostream& operator << ( std::ostream& out, const force_effect& f );

#endif


